@@ -12,20 +12,35 @@
 #define msgpack_pack_inline_func_cint(name) \
     static inline void msgpack_pack ## name
 
+// serialization context
 typedef struct {
     char *cur;       /* SvPVX (sv) + current output position */
     const char *end; /* SvEND (sv) */
     SV *sv;          /* result scalar */
+
+    bool prefer_int;
+    bool canonical;
 } enc_t;
 
-STATIC_INLINE void need(enc_t* const enc, STRLEN const len);
+STATIC_INLINE void
+dmp_append_buf(enc_t* const enc, const void* const buf, STRLEN const len)
+{
+    if (enc->cur + len >= enc->end) {
+        dTHX;
+        STRLEN const cur = enc->cur - SvPVX_const(enc->sv);
+        sv_grow (enc->sv, cur + (len < (cur >> 2) ? cur >> 2 : len) + 1);
+        enc->cur = SvPVX_mutable(enc->sv) + cur;
+        enc->end = SvPVX_const(enc->sv) + SvLEN (enc->sv) - 1;
+    }
+
+    memcpy(enc->cur, buf, len);
+    enc->cur += len;
+}
 
 #define msgpack_pack_user enc_t*
 
 #define msgpack_pack_append_buffer(enc, buf, len) \
-    need(enc, len);                               \
-    memcpy(enc->cur, buf, len);                   \
-    enc->cur += len;
+            dmp_append_buf(enc, buf, len)
 
 #include "msgpack/pack_template.h"
 
@@ -46,33 +61,32 @@ STATIC_INLINE void need(enc_t* const enc, STRLEN const len);
 
 #define ERR_NESTING_EXCEEDED "perl structure exceeds maximum nesting level (max_depth set too low?)"
 
+#define DMP_PREF_INT  "PreferInteger"
 
-STATIC_INLINE void need(enc_t* const enc, STRLEN const len)
-{
-    if (enc->cur + len >= enc->end) {
-        dTHX;
-        STRLEN const cur = enc->cur - SvPVX_const(enc->sv);
-        sv_grow (enc->sv, cur + (len < (cur >> 2) ? cur >> 2 : len) + 1);
-        enc->cur = SvPVX_mutable(enc->sv) + cur;
-        enc->end = SvPVX_const(enc->sv) + SvLEN (enc->sv) - 1;
+/* interpreter global variables */
+#define MY_CXT_KEY "Data::MessagePack::_pack_guts" XS_VERSION
+typedef struct {
+    bool prefer_int;
+    bool canonical;
+} my_cxt_t;
+START_MY_CXT
+
+
+static int dmp_config_set(pTHX_ SV* sv, MAGIC* mg) {
+    dMY_CXT;
+    assert(mg->mg_ptr);
+    if(strEQ(mg->mg_ptr, DMP_PREF_INT)) {
+        MY_CXT.prefer_int = SvTRUE(sv) ? true : false;
     }
-}
-
-
-static int s_pref_int = 0;
-
-STATIC_INLINE int pref_int_set(pTHX_ SV* sv, MAGIC* mg PERL_UNUSED_DECL) {
-    if (SvTRUE(sv)) {
-        s_pref_int = 1;
-    } else {
-        s_pref_int = 0;
+    else {
+        assert(0);
     }
     return 0;
 }
 
-MGVTBL pref_int_vtbl = {
+MGVTBL dmp_config_vtbl = {
     NULL,
-    pref_int_set,
+    dmp_config_set,
     NULL,
     NULL,
     NULL,
@@ -84,8 +98,18 @@ MGVTBL pref_int_vtbl = {
 };
 
 void init_Data__MessagePack_pack(pTHX_ bool const cloning) {
-    SV* var = get_sv("Data::MessagePack::PreferInteger", 0);
-    sv_magicext(var, NULL, PERL_MAGIC_ext, &pref_int_vtbl, NULL, 0);
+    if(!cloning) {
+        MY_CXT_INIT;
+        MY_CXT.prefer_int = false;
+        MY_CXT.canonical  = false;
+    }
+    else {
+        MY_CXT_CLONE;
+    }
+
+    SV* var = get_sv("Data::MessagePack::" DMP_PREF_INT, GV_ADDMULTI);
+    sv_magicext(var, NULL, PERL_MAGIC_ext, &dmp_config_vtbl,
+            DMP_PREF_INT, 0);
     SvSETMAGIC(var);
 }
 
@@ -144,7 +168,7 @@ STATIC_INLINE int try_int(enc_t* enc, const char *p, size_t len) {
 }
 
 
-STATIC_INLINE void _msgpack_pack_rv(enc_t *enc, SV* sv, int depth);
+STATIC_INLINE void _msgpack_pack_rv(pTHX_ enc_t *enc, SV* sv, int depth);
 
 STATIC_INLINE void _msgpack_pack_sv(enc_t* const enc, SV* const sv, int const depth) {
     dTHX;
@@ -156,7 +180,7 @@ STATIC_INLINE void _msgpack_pack_sv(enc_t* const enc, SV* const sv, int const de
         STRLEN const len     = SvCUR(sv);
         const char* const pv = SvPVX_const(sv);
 
-        if (s_pref_int && try_int(enc, pv, len)) {
+        if (enc->prefer_int && try_int(enc, pv, len)) {
             return;
         } else {
             msgpack_pack_raw(enc, len);
@@ -174,7 +198,7 @@ STATIC_INLINE void _msgpack_pack_sv(enc_t* const enc, SV* const sv, int const de
             msgpack_pack_double(enc, (double)SvNVX(sv));
         }
     } else if (SvROK(sv)) {
-        _msgpack_pack_rv(enc, SvRV(sv), depth-1);
+        _msgpack_pack_rv(aTHX_ enc, SvRV(sv), depth-1);
     } else if (!SvOK(sv)) {
         msgpack_pack_nil(enc);
     } else if (isGV(sv)) {
@@ -185,9 +209,14 @@ STATIC_INLINE void _msgpack_pack_sv(enc_t* const enc, SV* const sv, int const de
     }
 }
 
-STATIC_INLINE void _msgpack_pack_rv(enc_t *enc, SV* sv, int depth) {
+STATIC_INLINE
+void _msgpack_pack_he(pTHX_ enc_t* enc, HV* hv, HE* he, int depth) {
+    _msgpack_pack_sv(enc, hv_iterkeysv(he),   depth);
+    _msgpack_pack_sv(enc, hv_iterval(hv, he), depth);
+}
+
+STATIC_INLINE void _msgpack_pack_rv(pTHX_ enc_t *enc, SV* sv, int depth) {
     svtype svt;
-    dTHX;
     assert(sv);
     SvGETMAGIC(sv);
     svt = SvTYPE(sv);
@@ -211,9 +240,28 @@ STATIC_INLINE void _msgpack_pack_rv(enc_t *enc, SV* sv, int depth) {
 
         msgpack_pack_map(enc, count);
 
-        while ((he = hv_iternext(hval))) {
-            _msgpack_pack_sv(enc, hv_iterkeysv(he), depth);
-            _msgpack_pack_sv(enc, HeVAL(he), depth);
+        if (enc->canonical) {
+            AV* const keys = newAV();
+            sv_2mortal((SV*)keys);
+            av_extend(keys, count);
+
+            while ((he = hv_iternext(hval))) {
+                av_push(keys, SvREFCNT_inc(hv_iterkeysv(he)));
+            }
+
+            int const len = av_len(keys) + 1;
+            sortsv(AvARRAY(keys), len, Perl_sv_cmp);
+
+            int i;
+            for (i=0; i<len; i++) {
+                SV* sv = *av_fetch(keys, i, TRUE);
+                he = hv_fetch_ent(hval, sv, FALSE, 0U);
+                _msgpack_pack_he(aTHX_ enc, hval, he, depth);
+            }
+        } else {
+            while ((he = hv_iternext(hval))) {
+                _msgpack_pack_he(aTHX_ enc, hval, he, depth);
+            }
         }
     } else if (svt == SVt_PVAV) {
         AV* ary = (AV*)sv;
@@ -233,9 +281,9 @@ STATIC_INLINE void _msgpack_pack_rv(enc_t *enc, SV* sv, int depth) {
         char *pv = svt ? SvPV (sv, len) : 0;
 
         if (len == 1 && *pv == '1')
-            msgpack_pack_true(enc); 
+            msgpack_pack_true(enc);
         else if (len == 1 && *pv == '0')
-            msgpack_pack_false(enc); 
+            msgpack_pack_false(enc);
         else {
             //sv_dump(sv);
             croak("cannot encode reference to scalar '%s' unless the scalar is 0 or 1",
@@ -253,9 +301,10 @@ XS(xs_pack) {
         Perl_croak(aTHX_ "Usage: Data::MessagePack->pack($dat [,$max_depth])");
     }
 
+    SV* self  = ST(0);
     SV* val   = ST(1);
     int depth = 512;
-    if (items >= 3) depth = SvIV(ST(2));
+    if (items >= 3) depth = SvIVx(ST(2));
 
     enc_t enc;
     enc.sv        = sv_2mortal(newSV(INIT_SIZE));
@@ -263,6 +312,24 @@ XS(xs_pack) {
     enc.end       = SvEND(enc.sv);
     SvPOK_only(enc.sv);
 
+    // setup configuration
+    dMY_CXT;
+    enc.prefer_int = MY_CXT.prefer_int; // back compat
+    if(SvROK(self) && SvTYPE(SvRV(self)) == SVt_PVHV) {
+        HV* const hv = (HV*)SvRV(self);
+        SV** svp;
+
+        svp = hv_fetchs(hv, "prefer_integer", FALSE);
+        if(svp) {
+            enc.prefer_int = SvTRUE(*svp) ? true : false;
+        }
+
+        svp = hv_fetchs(hv, "canonical", FALSE);
+        if(svp) {
+            enc.canonical = SvTRUE(*svp) ? true : false;
+        }
+    }
+
     _msgpack_pack_sv(&enc, val, depth);
 
     SvCUR_set(enc.sv, enc.cur - SvPVX (enc.sv));
